import json
import logging
import requests
import traceback
import re
import socket
import _thread
from multiprocessing.pool import ThreadPool
import time
import ipaddress
import xml.etree.ElementTree as ET
from requests.auth import HTTPBasicAuth

import ruxit.api.selectors
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.data import PluginMeasurement, PluginProperty, MEAttribute
from ruxit.api.exceptions import AuthException
from ruxit.api.events import Event, EventMetadata
import urllib3

logger = logging.getLogger(__name__)

class CustomDataPowerPluginRemote(RemoteBasePlugin):
    # Used for threading
    start = 0
    aLock = _thread.allocate_lock()
    
    # Used for checking if the device has rebooted
    lastUptime = 0
    lastUptimeString = ""

    # Used for not spamming events
    previousEvents = []
    raisedEvents = []
    
    # Used for limiting events
    MAX_EVENTS = 99
    events = 0
    
    # For adding host IP as a listening IP
    listeningPorts = set()
    listeningIPs = set()
    
    # Used to always have debug logging enabled the first run
    firstRun = True
    
    def query(self, **kwargs):
        self.start = time.time()
        self.listeningPorts = set()
        self.listeningIPs = set()
        self.events = 0
        self.activeThreads = 0
        self.raisedEvents = []
        config = kwargs["config"]
        url = config["url"].strip()
        group = config["group"].strip()
        if "ips" in config:
            ips = config["ips"]
            if ips.strip() != "":
                for ip in ips.split(","):
                    self.listeningIPs.add(ip.strip())
        
        debugLogging = config["debug"]
        if debugLogging or self.firstRun:
            logger.setLevel(logging.DEBUG)
            self.firstRun = False
        else:
            logger.setLevel(logging.WARNING)
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
            
        # Compile the regexes used to find out which domains and services needs to be polled
        regexDomains = []
        if "domains" in config and config["domains"] != "":
            for domain in config["domains"].strip().split(','):
                regexDomains.append(re.compile(domain + "$"))
                logger.info("Adding domain filter: " + domain)
            
        regexServices = []
        pollServices = True
        if "services" in config and config["services"] != "":
            if config["services"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollServices = False
                logger.info("Not polling services")
            else:
                for service in config["services"].strip().split(','):
                    regexServices.append(re.compile(service + "$"))
                    logger.info("Adding service filter: " + service)
            
        regexInterfaces = []
        pollInterfaces = True
        if "interfaces" in config and config["interfaces"] != "":
            if config["interfaces"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollInterfaces = False
                logger.info("Not polling interfaces")
            else:
                for interface in config["interfaces"].strip().split(','):
                    regexInterfaces.append(re.compile(interface + "$"))
                    logger.info("Adding interface filter: " + interface)
            
        regexTasks = []
        pollTasks = True
        if "tasks" in config and config["tasks"] != "":
            if config["tasks"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollTasks = False
                logger.info("Not polling tasks")
            else:
                for task in config["tasks"].strip().split(','):
                    regexTasks.append(re.compile(task + "$"))
                    logger.info("Adding task filter: " + task)
            
        regexXmlMgrs = []
        pollXmlMgrs = True
        if "xmlMgrs" in config and config["xmlMgrs"] != "":
            if config["xmlMgrs"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollXmlMgrs = False
                logger.info("Not polling xmlMgrs")
            else:
                for xmlMgr in config["xmlMgrs"].strip().split(','):
                    regexXmlMgrs.append(re.compile(xmlMgr + "$"))
                    logger.info("Adding xmlMgr filter: " + xmlMgr)
            
        regexLogTargets = []
        pollLogTargets = True
        if "logTargets" in config and config["logTargets"] != "":
            if config["logTargets"].replace(",", "").replace("'", "").strip().lower() == "none":
                pollLogTargets = False
                logger.info("Not polling logTargets")
            else:
                for logTarget in config["logTargets"].strip().split(','):
                    regexLogTargets.append(re.compile(logTarget + "$"))
                    logger.info("Adding logTarget filter: " + logTarget)
        
        hostName = url.replace("http://","").replace("https://","")
        hostName = hostName.split(':')[0].split('/')[0]
        g1 = self.topology_builder.create_group("DataPower - " + group, "DataPower - " + group)
        e1 = g1.create_element("DataPower - " + group + " - " + hostName, "DataPower - " + group + " - " + hostName)
        e1.report_property("Plugin version", kwargs["json_config"]["version"])
        
        # All the metrics currently polled
        # name = name of the metric as defined in the plugin.json
        # field = name of the field in the returned xml where the metric value is written
        # class = name of the class which needs to be polled in the post body of the xml request
        # metric = absolute or relative
        # splitByDomain = is it a device or a domain metric
        # limitSplitting = only include splitting if it matches a regex in the passed array
        # expectedValue = if it is a string value instead of a number
        # expectedLowerThan = value needs to be smaller than the passed field
        # expectedGreaterThan = value needs to be greater than the passed field
        # splitName = the dimension name as defined in the plugin.json
        # split = the field which contains the value to split by
        # useClassForEventElement = If there is an expected value, either use the class name, or an event group/event element
        # eventGroup = Only used when useClassForEventElement is False. Optional, used for displaying the content of another element in the name of the event
        # eventElement = Required when useClassForEventElement is False, used for displaying the content of another element as the name of the event
        # filters = an array of AND filters, each containing the field which needs to contain one of the provided values

        metrics = []
        metrics.append({"name": "cryptoEngineStatus", "field": "EngineStatus", "class": "CryptoEngineStatus", "metric": "absolute", "splitByDomain": False, "expectedValue": "fully operational", "useClassForEventElement": True})
        metrics.append({"name": "activeSubscribers", "field": "SpoolerCount", "class": "WSMAgentStatus", "metric": "absolute", "splitByDomain": True})
        metrics.append({"name": "recordsLost", "field": "RecordsLost", "class": "WSMAgentStatus", "metric": "relative", "splitByDomain": True})
        metrics.append({"name": "pendingRecordsQueue", "field": "PendingRecords", "class": "WSMAgentStatus", "metric": "absolute", "splitByDomain": True})
        metrics.append({"name": "completeRecordsQueue", "field": "CompleteRecords", "class": "WSMAgentStatus", "metric": "absolute", "splitByDomain": True})
        metrics.append({"name": "Disk.freeEncrypted", "field": "FreeEncrypted", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Disk.totalEncrypted", "field": "TotalEncrypted", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Disk.freeTemporary", "field": "FreeTemporary", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Disk.totalTemporary", "field": "TotalTemporary", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Disk.freeInternal", "field": "FreeInternal", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Disk.totalInternal", "field": "TotalInternal", "class": "FilesystemStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "cpuload", "field": "oneMinute", "class": "CPUUsage", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Memory.Usage", "field": "Usage", "class": "MemoryStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Memory.Total", "field": "TotalMemory", "class": "MemoryStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Memory.Used", "field": "UsedMemory", "class": "MemoryStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Memory.Free", "field": "FreeMemory", "class": "MemoryStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "Memory.Req", "field": "ReqMemory", "class": "MemoryStatus", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "load", "field": "Load", "class": "SystemUsage", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "worklist", "field": "WorkList", "class": "SystemUsage", "metric": "absolute", "splitByDomain": False})
        metrics.append({"name": "temperatureWarning", "field": "Value", "split": "Name", "splitName": "Sensor", "class": "TemperatureSensors", "metric": "none", "splitByDomain": False, "expectedLowerThan": "UpperCriticalThreshold", "useClassForEventElement": False, "eventElement": "Name"})
        metrics.append({"name": "fanWarning", "field": "FanSpeed", "split": "FanID", "splitName": "Fan", "class": "EnvironmentalFanSensors", "metric": "none", "splitByDomain": False, "expectedHigherThan": "LowerCriticalThreshold", "useClassForEventElement": False, "eventElement": "FanID"})
        metrics.append({"name": "powersupplyWarning", "field": "powersupply", "class": "EnvironmentalSensors", "metric": "none", "splitByDomain": False, "expectedValue": "ok", "eventElementText": "PowerSupply"})
        metrics.append({"name": "connectionsAccepted", "field": "oneMinute", "class": "ConnectionsAccepted", "metric": "absolute", "splitByDomain": False})
        if pollServices:
            metrics.append({"name": "Service.httpTransactionThroughput", "field": "oneMinute", "split": "proxy", "splitName": "Service", "class": "HTTPTransactions", "metric": "absolute", "splitByDomain": True, "limitSplitting": regexServices, "multiply": 60})
            metrics.append({"name": "Service.httpTransactionsTime", "field": "oneMinute", "split": "proxy", "splitName": "Service", "class": "HTTPMeanTransactionTime", "metric": "absolute", "splitByDomain": True, "limitSplitting": regexServices})
            metrics.append({"name": "Service.status", "field": "OpState", "split": "Name", "splitName": "Service", "class": "ObjectStatus", "metric": "absolute", "splitByDomain": True, "expectedValue": "up", "limitSplitting": regexServices, "useClassForEventElement": False, "eventGroup": "Class", "eventElement": "Name", "filters": [{"field": "AdminState", "valueIn": ["enabled"]}, {"field": "Class", "valueIn": ["HTTPService", "MultiProtocolGateway", "SSLProxyService", "TCPProxyService", "WebAppFW", "WebTokenService", "WSGateway", "XMLFirewallService", "XSLCoprocService", "XSLProxyService"]}]})
        if pollInterfaces:
            metrics.append({"name": "Interface.receiveBps", "field": "RxHCBytes", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.receivePackets", "field": "RxHCPackets", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.receiveErrors", "field": "RxErrors2", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.receiveDrops", "field": "RxDrops2", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.transmitBps", "field": "TxHCBytes", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.transmitPackets", "field": "TxHCPackets", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.transmitErrors", "field": "TxErrors2", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
            metrics.append({"name": "Interface.transmitDrops", "field": "TxDrops2", "split": "IP", "splitName": "Interface", "class": "EthernetInterfaceStatus", "metric": "relative", "splitByDomain": False, "limitSplitting": regexInterfaces})
        if pollTasks:
            metrics.append({"name": "Task.load", "field": "Load", "split": "TaskName", "splitName": "Task", "class": "SystemUsage2Table", "metric": "absolute", "splitByDomain": False, "limitSplitting": regexTasks})
            metrics.append({"name": "Task.workList", "field": "WorkList", "split": "TaskName", "splitName": "Task", "class": "SystemUsage2Table", "metric": "absolute", "splitByDomain": False, "limitSplitting": regexTasks})
            metrics.append({"name": "Task.CPU", "field": "CPU", "split": "TaskName", "splitName": "Task", "class": "SystemUsage2Table", "metric": "absolute", "splitByDomain": False, "limitSplitting": regexTasks})
            metrics.append({"name": "Task.Memory", "field": "Memory", "split": "TaskName", "splitName": "Task", "class": "SystemUsage2Table", "metric": "absolute", "splitByDomain": False, "limitSplitting": regexTasks})
            metrics.append({"name": "Task.FileCount", "field": "FileCount", "split": "TaskName", "splitName": "Task", "class": "SystemUsage2Table", "metric": "absolute", "splitByDomain": False, "limitSplitting": regexTasks})
        if pollXmlMgrs:
            metrics.append({"name": "XMLManager.Requests", "field": "reqOneMin", "split": "XMLManager", "splitName": "XMLManager", "class": "HTTPConnections", "metric": "absolute", "splitByDomain": True, "limitSplitting": regexXmlMgrs})
        if pollLogTargets:
            metrics.append({"name": "LogTarget.process", "field": "EventsProcessed", "split": "LogTarget", "splitName": "LogTarget", "class": "LogTargetStatus", "metric": "relative", "splitByDomain": True, "limitSplitting": regexLogTargets})
            metrics.append({"name": "LogTarget.drop", "field": "EventsDropped", "split": "LogTarget", "splitName": "LogTarget", "class": "LogTargetStatus", "metric": "relative", "splitByDomain": True, "limitSplitting": regexLogTargets})
            metrics.append({"name": "LogTarget.pending", "field": "EventsPending", "split": "LogTarget", "splitName": "LogTarget", "class": "LogTargetStatus", "metric": "absolute", "splitByDomain": True, "limitSplitting": regexLogTargets})
        
        # The initial poll classes
        initialClasses = []
        initialClasses.append("DomainStatus") # The names of all of the domains
        initialClasses.append("IPAddressStatus") # The IP addresses of the device
        
        # Compiling a list of the classes for metrics and parameters
        perDeviceClasses = []
        perDeviceClasses.append("DateTimeStatus") # The uptime, used for restart events
        perDeviceClasses.append("FirmwareStatus") # Device properties
        perDomainClasses = []
        perDomainClasses.append("ServicesStatus") # The ports used by the domain
        
        for metric in metrics:
            if metric["splitByDomain"]:
                if metric["class"] not in perDomainClasses:
                    perDomainClasses.append(metric["class"])
            else:
                if metric["class"] not in perDeviceClasses:
                    perDeviceClasses.append(metric["class"])
                
        root = self.getRESTResult("default", initialClasses, config, e1)
        if root == False: # Polling failed
            e1.absolute(key = "connectivity", value = 0)
            self.report_custom_info_event(e1, "Connection to " + hostName + " failed", "Connectivity issue")
        else:
            e1.absolute(key = "connectivity", value = 100)
            root_map = self.getRootMap(root) # A key - value object linking the classes to the array index
            parent_map = self.getParentMap(root) # A key - value object linking all nodes to their parent
            
            # Create a list of the domains which should be polled
            matchedDomains = []
            try:
                for elem in root[root_map["DomainStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/DomainStatus'):
                    domain = elem.find('Domain').text
                    logger.info("Found domain: " + domain)
                    if len(regexDomains) == 0:
                        matchedDomains.append(domain)
                        logger.info("Matching domain due to 0 size filter: " + domain)
                    else:
                        for regex in regexDomains:
                            if regex.match(domain):
                                matchedDomains.append(domain)
                                logger.info("Matching domain due to regex '" + str(regex) + "': " + domain)
            except Exception as e:
                logger.exception(e)
                raise AuthException("0 domains returned for user '" + config['auth_user'] + "', is the user/profile initialized?")
                
            try:
                for elem in root[root_map["IPAddressStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/IPAddressStatus'):
                    ipVersion = elem.find('IPVersion').text
                    if ipVersion == "ipv4":
                        ipAddress = elem.find('IP').text
                        self.listeningIPs.add(ipAddress)
            except Exception as e:
                logger.info("No parsable IPAddressStatus returned")
                    
            pool = ThreadPool(processes = 10)
            
            for domain in matchedDomains:
                logger.info("Polling " + domain)
                pool.apply_async(self.callDataPower, (True, domain, perDomainClasses, config, metrics, e1))
                    
            for className in perDeviceClasses:
                logger.info("Polling default, class: " + className)
                pool.apply_async(self.callDataPower, (False, "default", [className], config, metrics, e1))
                
            pool.close()
            
            # Max polling time is 50 seconds
            process_alive = 0
            while time.time() - self.start < 50:
                process_alive = 0
                for process in pool._pool:
                    if process.is_alive():
                        process_alive += 1
                if process_alive == 0:
                    logger.info('All processes finished, exiting checking loop')
                    break
                time.sleep(1)
            
            # If we get to this point, a process was flagged alive and it's been more than 50 seconds, then terminate it
            # Otherwise, if process_alive was not set above, then no process should be alive at this point and no need to terminate
            if process_alive > 0:
                pool.terminate()
                pool.join()
                logger.warning(str(process_alive) + ' processes are alive. Terminating before finishing polling cycle')
            

            # If calling the XML Management Interface on an IP, add it as a communication ip
            hostIP = None
            try:
                socket.inet_aton(hostName) # is the host name an ip
                hostIP = hostName
            except socket.error as e:
                try:
                    hostIP = socket.gethostbyname(hostName)
                except socket.error as e:
                    logger.exception(e)
            if hostIP is not None:
                self.listeningIPs.add(hostIP)
            
            endpointsAssigned = 0
            for ip in self.listeningIPs:
                if ip != None and ip != "0.0.0.0" and ip != "127.0.0.1":
                    try:
                        if endpointsAssigned < 100:
                            e1.add_endpoint(ip)
                            logger.debug("Adding endpoint " + ip)
                            endpointsAssigned += 1
                        else:
                            logger.info("Too many endpoints, skipping " + ip)
                    except TypeError as e:
                        for port in self.listeningPorts:
                            if not isinstance(port, int) and port.isdigit():
                                port = int(port)
                            if port > 0 and port < 65536:
                                if endpointsAssigned < 100:
                                    e1.add_endpoint(ip, int(port))
                                    logger.debug("Adding endpoint " + ip + ":" + str(port))
                                    endpointsAssigned += 1
                                else:
                                    logger.info("Too many endpoints, skipping " + ip + ":" + str(port))

            # Register all metrics which contain a result
            for metric in metrics:
                if "splits" in metric:
                    for split in metric["splits"]:
                        if metric["metric"] == "absolute":
                            e1.absolute(key = metric["name"], value = split["result"], dimensions = split["dimensions"])
                        else:
                            e1.relative(key = metric["name"], value = split["result"], dimensions = split["dimensions"])
                elif "result" in metric:
                    if metric["metric"] == "absolute":
                        e1.absolute(key = metric["name"], value = metric["result"])
                    else:
                        e1.relative(key = metric["name"], value = metric["result"])
        self.previousEvents = self.raisedEvents
                    
    def callDataPower(self, domainMetrics, domain, classes, config, metrics, e1):
        try:
            requestStartTime = time.time()
            logger.info("Polling: " + domain)
            root = self.getRESTResult(domain.strip(), classes, config, e1)
            if root != False: # Polling successful
                root_map = self.getRootMap(root) # A key - value object linking the classes to the array index
                parent_map = self.getParentMap(root) # A key - value object linking all nodes to their parent
                    
                # Get the ports used by the domain
                if "ServicesStatus" in root_map:
                    for elem in root[root_map["ServicesStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/ServicesStatus/'):
                        if time.time() - self.start < 50:
                            port = int(next(iter(parent_map[elem].findall("LocalPort")), "").text)
                            ip = next(iter(parent_map[elem].findall("LocalIP")), "").text
                            self.listeningPorts.add(port)
                            if ip is not None:
                                self.listeningIPs.add(ip)
                                
                if "DateTimeStatus" in root_map:
                    try:
                        for elem in root[root_map["DateTimeStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/DateTimeStatus'):
                            uptime = elem.find("uptime2").text
                            uptimeParts = uptime.split(" ")
                            uptimeTimeParts = uptimeParts[2].split(":")
                            currentUpTime = int(uptimeTimeParts[2]) + int(uptimeTimeParts[1])*60 + int(uptimeTimeParts[0])*60*60 + int(uptimeParts[0])*60*60*24
                            with self.aLock:
                                if currentUpTime > 0:
                                    if self.lastUptime > currentUpTime:
                                        self.report_custom_info_event(e1, "The device was restarted after an uptime of " + self.lastUptimeString + ". The new uptime is " + uptime, "Device restarted")
                                        logger.info("DataPower restart detected. Old uptime: " + str(self.lastUptime) + "s. New uptime: " + str(currentUpTime) + "s")
                                    self.lastUptimeString = uptime
                                    self.lastUptime = currentUpTime
                                    e1.report_property("Uptime", uptime)
                    except Exception as e:
                        logger.exception("Failed getting the uptime")
                        logger.exception(e)
                        
                if "FirmwareStatus" in root_map:
                    with self.aLock:
                        for elem in root[root_map["FirmwareStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/FirmwareStatus'):
                            e1.report_property("Type", elem.find('Type').text)
                            e1.report_property("Version", elem.find('Version').text)
                            e1.report_property("Build", elem.find('Build').text)
                
                if "EthernetInterfaceStatus" in root_map:
                    with self.aLock:
                        for elem in root[root_map["EthernetInterfaceStatus"]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/EthernetInterfaceStatus'):
                            ethernetIP = elem.find('IP').text
                            if ethernetIP is not None:
                                if "/" not in ethernetIP:
                                    self.listeningIPs.add(ethernetIP)
                                elif int(ethernetIP.split("/")[1]) >= 30:
                                    net4 = ipaddress.ip_network(ethernetIP, False)
                                    self.listeningIPs.add(str(next(net4.hosts())))
                    
                # Add the results of the domain specific metrics to the metrics object
                for metric in metrics:
                    if metric["splitByDomain"] == domainMetrics and metric["class"] in classes:
                        with self.aLock:
                            if time.time() - self.start < 49:
                                self.updateMetric(root, root_map, metric, parent_map, domain, e1)
            logger.info("Domain: " + domain + ", class(es): " + str(classes) + ", performance: " + str(time.time() - requestStartTime) + "s")
        except Exception as e:
            logger.info("Failed polling domain: " + str(domain) + ", class(es): " + str(classes))
            logger.exception(e)
        with self.aLock:
            self.callsToGo -= 1
            self.activeThreads -= 1
    
    # Poll the XML Management interface of the DataPower device
    def getRESTResult(self, domain, classes, config, e1):
        auth_user = config['auth_user']
        auth_password = config['auth_password']
        
        #create verify request parameter
        verifyCertificate = config["certificate"]
        certificatePath = config["certificatePath"].strip()
        clientCert = False
        if "clientCert" in config:
            clientCert = config["clientCert"]
        clientCertFile = ""
        if "clientCertFile" in config:
            clientCertFile = config["clientCertFile"].strip()
        clientCertKey = ""
        if "clientCertKey" in config:
            clientCertKey = config["clientCertKey"].strip()
			
        verify = verifyCertificate
        if verify and certificatePath != "":
            verify = certificatePath
            
        cert = None
        if clientCert and clientCertFile != "":
            if clientCertKey != "":
                cert = (clientCertFile, clientCertKey)
            else:
                cert = clientCertFile
            
        url = config['url']
        headers = {'content-type': 'text/xml'}
        body = '<?xml version="1.0" encoding="UTF-8"?><env:Envelope xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"><env:Body>'
        # Generate the post body, taking into account if the device or the domain needs to be queried
        for className in classes:
            body += '<dp:request domain="' + domain + '" xmlns:dp="http://www.datapower.com/schemas/management"><dp:get-status class="' + className + '"/></dp:request>'
        body += '</env:Body></env:Envelope>'
        # Generates a lot of logging, use on demand
        # logger.info("PostBody: " + body)

        basicAuth = None; # Should we use basic authentication?
        if 'auth_user' in config and 'auth_password' in config:
            if auth_user or auth_password:
                basicAuth = HTTPBasicAuth(auth_user, auth_password)
            
        try:
            response = requests.post(url, auth=basicAuth, data=body, headers=headers, cert=cert, verify=verify, timeout=50 - (time.time() - self.start))
        except Exception as e:
            logger.warning("Failed polling: " + url + " with domain: " + domain)
            logger.exception(e)
            return False # The request failed
            
        content = response.text
            
        # Generates a lot of logging, use on demand
        # logger.info("Reply: " + content)
        
        if response.status_code != 200:
            logger.exception("Failed polling URL: " + url + " with domain: " + domain + ". Status code: " + str(response.status_code) + ". Request: " + body + ". Reply: " + content)
            self.report_custom_info_event(e1, "Failed polling domain '" + domain + "'. Status code: " + str(response.status_code), "Polling issue")
            return False # The request failed
        
        if "Authentication failure" in content or "Access Denied" in content:
            raise AuthException("Authentication failed for user '" + auth_user + "'")
            
        return ET.fromstring(re.sub(r"(<\?xml[^>]+\?>)", r"\1<root>", content) + "</root>") # Add a fake root element, the reply from DataPower is not correctly formatted XML
    
    # Returns a key - value object linking the classes to the array index
    def getRootMap(self, root):
        root_map = {};
        root_id = 0
        for rootObj in root:
            for elem in rootObj.findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status'):
                for child in elem:
                    root_map[child.tag] = root_id
            root_id += 1
        return root_map
        
    # Get the metric value from the xml body and store it in the metric object
    def updateMetric(self, root, root_map, metric, parent_map, domain, e1):
        try:
            for elem in root[root_map[metric["class"]]].findall('{http://schemas.xmlsoap.org/soap/envelope/}Body/{http://www.datapower.com/schemas/management}response/{http://www.datapower.com/schemas/management}status/' + metric["class"] + '/' + metric["field"]):
                getMetric = False
                if "filters" in metric: # If there are filters to only get a sub-set of the values
                    parent = parent_map[elem]
                    filtersToPass = len(metric["filters"])
                    for filter in metric["filters"]:
                        for filterField in parent_map[elem].findall(filter["field"]):
                            for valueIn in filter["valueIn"]:
                                if filterField.text == valueIn:
                                    filtersToPass = filtersToPass - 1
                                    if filtersToPass == 0:
                                        getMetric = True
                else:
                    getMetric = True
                if getMetric:
                    # Used for all "expected" methods to get an event name
                    eventObject = metric["class"]
                    if "eventElementText" in metric:
                        eventObject = metric["eventElementText"]
                    elif "eventElement" in metric and not metric["useClassForEventElement"]:
                        eventObject = ""
                        if "eventGroup" in metric:
                            eventObject = next(iter(parent_map[elem].findall(metric["eventGroup"])), None).text + " - "
                        eventObject = eventObject + next(iter(parent_map[elem].findall(metric["eventElement"])), None).text
                    if "expectedLowerThan" in metric: # If the metric needs to be lower than another field
                        lowerThan = next(iter(parent_map[elem].findall(metric["expectedLowerThan"])), None).text
                        if lowerThan.isnumeric() and elem.text.isnumeric():
                            if int(lowerThan) <= int(elem.text):
                                self.report_custom_info_event(e1, eventObject + " expected result: lower than " + lowerThan + ", received: " + elem.text, eventObject + " - failure")
                    if "expectedHigherThan" in metric: # If the metric needs to be higher than another field
                        higherThan = next(iter(parent_map[elem].findall(metric["expectedHigherThan"])), None).text
                        if higherThan.isnumeric() and elem.text.isnumeric():
                            if int(higherThan) >= int(elem.text):
                                self.report_custom_info_event(e1, eventObject + " expected result: higher than " + lowerThan + ", received: " + elem.text, eventObject + " - failure")
                    if "expectedValue" in metric: # If the metric is a string instead of a number
                        if metric["expectedValue"] == elem.text:
                            value = 100
                        else:
                            value = 0
                            self.report_custom_info_event(e1, eventObject + " expected result: " + metric["expectedValue"] + ", received: " + elem.text, eventObject + " - failure")
                    else:
                        value = int(elem.text)
                    if "multiply" in metric:
                        value = value * metric["multiply"]
                        
                    if metric["metric"] != "none":
                        if "split" in metric:
                            for split in parent_map[elem].findall(metric["split"]):
                                if split.text != None:
                                    splitLimitationPassed = False
                                    if "limitSplitting" in metric:
                                        logger.info("Limiting splitting, text: " + split.text)
                                        if len(metric["limitSplitting"]) == 0:
                                            splitLimitationPassed = True
                                            logger.info("Matching splitting due to 0 size filter: " + split.text)
                                        else:
                                            for regex in metric["limitSplitting"]:
                                                if regex.match(split.text):
                                                    splitLimitationPassed = True
                                                    logger.info("Matching splitting due to regex '" + str(regex) + "': " + split.text)
                                                    break
                                    else:
                                        splitLimitationPassed = True
                                    if splitLimitationPassed:
                                        if not "splits" in metric:
                                            metric["splits"] = []
                                        dimensions = {}
                                        dimensions[metric["splitName"]] = split.text
                                        if metric["splitByDomain"] and domain != None:
                                            dimensions["Domain"] = domain
                                        if len(metric["splits"]) < 1000:
                                            metric["splits"].append({"result": value, "dimensions": dimensions})
                                        else:
                                            logger.info("Max 1000 splittings per metric reached for " + metric["name"] + ", ignoring metric with dimensions: " + str(dimensions))
                        else:
                            if metric["splitByDomain"] and domain != None:
                                dimensions = {}
                                dimensions["Domain"] = domain
                                if not "splits" in metric:
                                    metric["splits"] = []
                                if len(metric["splits"]) < 1000:
                                    metric["splits"].append({"result": value, "dimensions": dimensions})
                                else:
                                    logger.info("Max 1000 splittings per metric reached for " + metric["name"] + ", ignoring metric with dimensions: " + str(dimensions))
                            else:
                                if "result" in metric:
                                    metric["result"] += value
                                else:
                                    metric["result"] = value
        except KeyError as e:
            if domain is not None:
                logger.info('Domain: ' + domain + ', ' + metric["class"] + '/' + metric["field"] + " class/field key missing")
            else:
                logger.info(metric["class"] + '/' + metric["field"] + " class/field key missing")
        except Exception as e:
            if domain is not None:
                logger.exception('Domain: ' + domain + ', ' + metric["class"] + '/' + metric["field"] + " class/field failed")
            else:
                logger.exception(metric["class"] + '/' + metric["field"] + " class/field failed")
            logger.warn(traceback.format_exc())
                
    # Returns a key - value object linking all nodes to their parent
    def getParentMap(self, root):
        return {c:p for p in root.iter() for c in p}
        
    def report_custom_info_event(self, e1, message, title):
        if self.events < self.MAX_EVENTS:
            if title + message not in self.previousEvents:
                self.events += 1
                e1.report_custom_info_event(message, title)
            self.raisedEvents.append(title + message)